{$M 8192,0,655360}
{
Dumps a floppy diskette to an image file.  Written by trixter@oldskool.org

Standard/universal image file for raw/rawrite/wimage floppy disk images:
for tracks 0..N
  for heads 0..1
    for sectors 1..N
      (sector data)
t0h0s1,t0h0s2,t0h0s3...t0h1s1,t0h1s2,t0h1s2...t1h0s1,t1h0s2,t1h0s3... etc.

This imager has special handling for disks with bad sectors:
- It will try to re-read all bad sectors after the good tracks have been saved
- If multiple re-reads fail, the sector will be dumped as all "!!!"s to
  clearly indicate which sectors could not be read

The behavior of dumping "!!!"s for unreadable sectors is known as the "Better
the devil you know, than the devil you don't" strategy.  Please see the
DISK2IMG.TXT documentation for more information and additional options.
}

{{$DEFINE DEBUG}    {general debugging; undefine for production release}
{{$DEFINE NODISKIO} {for debugging; skips some disk I/O}
{{$DEFINE SANITY}   {function sanity checks}

Program disk2img;

uses
  objects,
  streams,
  d2i_globals, {global variables}
  d2i_io,      {low-level disk utilities}
  d2i_support, {additional support routines}
  support,     {misc support routines}
  TFastWrite,  {fast screen writing}
  TInterrupts; {interrupt and timer support}

const
  badsec9s:byte=0;  {quick'n'dirty format checking}
  badhead1s:word=0; {quick'n'dirty format checking}
  recoverCount:word=0;
  refineCount:word=0;
  refineLimit=5;    {finish refining once reconstructed sector is static this many times}
  Recover:boolean=false;
  TrackDump:boolean=true;


Procedure DumpInit;

  Procedure SetupVars;
  var
    status:word;
    tmpe:string[12];
    dsp:^boot_structure;
    _t,_h,_s:byte;

  begin
    {Init sector status array}
    fillchar(readStatus,sizeof(readstatus),unknown);
    getmem(sectorbuffer,sectorsize);
    getmem(tempsector,sectorsize);

    status:=ResetDrive(CurDrive);
    if (status<>0) and not ignoreErrors
      then fatalerror(1,'"'+DriveStatusString(status)+'" when resetting drive subsystem');

    if Query then begin
      status:=ReadSectors(CurDrive, 0, 0, 1, 1, sectorbuffer);
      if ((status and $00FF)<>1) {should get number of sectors read as return value}
        then fatalerror(2,'Nothing read; "'+DriveStatusString(GetDriveStatus(curdrive))+'" when trying to query boot sector');
      dsp:=sectorbuffer;
      if (dsp^.sectors_per_track = 0)
      or (dsp^.sectors_per_track > 18)
      or (dsp^.number_of_heads = 0)
      or (dsp^.number_of_heads > 2)
        then fatalerror(3,'BIOS Parameter Block bad; DOS-formatted diskette?  Re-run without /Q');
      numsectors:=dsp^.sectors_per_track;
      numheads:=dsp^.number_of_heads;
      numtracks:=dsp^.total_sectors div numsectors div numheads;
    end;

    tracksize:=sectorsize*numsectors*numheads;
    totalsectors:=numtracks*numsectors*numheads;
    imagesize:=longint(numtracks)*tracksize;
    totaltrackreads:=numtracks*numheads;
    getmemDMAPageAligned(trackbuffer,tracksize);

    if renameExt then begin
      tmpe:=inttostr(((numtracks*numsectors) div 2)*numheads); {make extension "360" or "1440" etc.}
      if length(tmpe) > 3 then byte(tmpe[0]):=3; {truncate extension if necessary}
      imageName:=basename(imageName)+'.'+tmpe;
    end;

    salvagedSectorp:=@salvagedSector;
    oldSectorp:=@oldSector;
  end;

  Procedure PrepDrive;
  var
    chkstatus:word;
    sec:byte;
  begin
    {initialize the floppy controller subsystem}
    {read dummy sector three times in a row to force drive to spin up}
    DisplayStatus('Forcing motor to spin...');
    for sec:=0 to 2 do begin
      DisplayStatus('Reading sector #0...');
      {read only first sector since that is the only sector "guaranteed" to be readable}
      chkstatus:=ReadSectors(CurDrive,0,0,1,1,sectorbuffer);
      if lo(chkstatus)<>1 {no sectors read} then begin
        DisplayStatus('Wrong number of sectors read; resetting drive...');
        resetdrive(CurDrive);
      end else break;
    end;

    {if we never read anything successfully, abort}
    if (lo(chkstatus)<>1) and not ignoreErrors
      then fatalerror(10,'All reads returned errors; you should reboot to reset NEC765 and BIOS/BDA');

    DisplayStatus('Checking drive motor...');
    {$IFNDEF NODISKIO}
    if (DriveMotorStatus and $F = 0) and not recover
      then fatalerror(5,'Disk drive did not spin up');
    {$ENDIF}
  end;

  Procedure PrepImages;
  var
    badstream:TNulstream;
    s:string;
    t:pointer;
  begin
    DisplayStatus('Creating temporary image...');
    imageTemp:=TempStream(imagesize,imagesize,ForSpeed);

    DisplayStatus('Creating image file...');
    new(imageFile,init(imageName,stCreate,TrackSize));

    {Fill temp image with !!!s to assume bad reads}
    s:='Initializing temporary image located ';
    t:=typeof(imageTemp^);
         if (t=typeof(TEMSStream)) or (t=typeof(TEMSStream2))
                                        then s:=s+'in EMS'
    else if (t=typeof(TNamedBufStream)) or (t=typeof(TTempBufStream))
                                        then s:=s+'on disk'
    else if t = typeof(TMemoryStream)   then s:=s+'in lower DOS RAM'
    else if t = typeof(TXMSStream)      then s:=s+'in XMS'
    else s:=s+'in a sea of ones and zeros';
    s:=s+'...';
    DisplayStatus(s);
    badstream.init(ord('!'));
    FastCopy(badstream,imagetemp^,imagesize);
    badstream.done;

    if (imageTemp=nil)
    or (imageFile=nil)
    or (imageTemp^.Status <> stOK)
    or (imageFile^.Status <> stOK)
      then fatalerror(81,'Something went wrong prepping images');
  end;

begin
  PrintWelcome;
  GetOptions;
  SetupVars;
  SetupScreen;
  PrepImages;
  PrepDrive;
end;

{$IFDEF NODISKIO}
Function fakeStatus:word;
begin
  fakeStatus:=(random(2) SHL 8) OR (random(numSectors)+1)
end;
{$ENDIF}

Procedure DumpDo(mode:phase);

var
  curstatus:word;
  ch:char;
  sector,head,track:byte;
  destp:^byte;
  status:byte;
  tr,se,he:byte;
  tmps:string;
  l:longint;
  b:byte;
  readmatrixdelay:longint;
  oldLSOT:byte;

  Function AbsPos(_track,_head,_sec:word):longint;
  var
    result:longint;
  begin
    dec(_sec);
    result:=(longint(_track) * longint(tracksize))
          + (longint(_head) * (tracksize div 2))
          + (longint(_sec) * sectorsize);
    AbsPos:=result;
  end;

  Procedure CommitTrack(trackbuf:pointer;tracknum:word;headnum:byte);
  begin
    (*abspos:=longint(tracksize) * longint(tracknum); {casting to work around arithmetic overflow}*)
    imageTemp^.seek(abspos(tracknum,headnum,1));
    imageTemp^.write(trackBuf^,tracksize div numheads);
  end;

  Procedure CommitSector(secbuf:pointer;tracknum,headnum,sectnum:word);
  begin
    imageTemp^.seek(abspos(tracknum,headnum,sectnum));
    imageTemp^.write(secbuf^,sectorsize);
  end;

begin
  curstatus:=0;
  numErrors:=0;

  {If initial run, we dump all tracks quickly.  If a second run is needed
  because we got errors, we attempt to dump only sectors that had problems.}

  if mode=initial then begin
    trackDump:=true;
    Recover:=false;
   end else begin
    trackDump:=false;
    Recover:=true;
  end;

  {Set the DBT to match what we're trying to read just in case DOS didn't}
  oldLSOT:=DBT^.LastSectorOnTrack;
  DBT^.LastSectorOnTrack:=numsectors;
  ResetDrive(CurDrive);

  {calc # of sectors left from read status matrix; used for % time left}
  starttick:=TicksSinceMidnight;
  sectorcount:=0;
  sectorsRemaining:=0;
  for track:=0 to numtracks-1 do
    for head:=0 to numheads-1 do
      for sector:=1 to numsectors do
        if readStatus[track,head,sector] <> good
          then inc(sectorsRemaining);

  fillchar(trackbuffer^,tracksize,0);
  for track:=0 to numtracks-1 do begin
    {reset destintation to start of track buffer}
    destp:=trackbuffer;
    {Repaint read matrix on even tracks}
    if trackdump
      then if (track AND 7)=0
        then DisplayReadMatrix;
    for head:=0 to numheads-1 do begin
      {fill a track buffer (both heads) before writing to file and moving on}
      if trackdump then begin
        sector:=1; {we're reading an entire track so start at sector 1}
        {status returned in AH, num sectors read in AL}
        DisplayDriveStatus(track,head,0,false); {0=doing a track dump}
        curstatus:=ReadSectors(CurDrive,head,track,sector,numsectors,destp);
        {$IFDEF NODISKIO}
        curstatus:=fakestatus;
        move(ptr(0,0)^,trackbuffer^,tracksize);
        {$ENDIF}
        DisplaySector(destp,sectorline);
        tmps:='Last status code: '+DriveStatusString(hi(curstatus));
        DisplayStatus(tmps);

        if hi(curstatus)<>0 then begin
          {bad read; inc error count and mark track as bad}
          inc(numErrors);
          {mark track as bad}
          for se:=1 to numSectors do readStatus[track,head,se]:=bad;
        end else begin
          {good read for this head; mark track as good, commit to image}
          CommitTrack(destp,track,head);
          for se:=1 to numSectors do readStatus[track,head,se]:=good;
        end;

        {increment dest. pointer if we have more heads to read}
        inc(longint(destp),tracksize div numheads);
        inc(trackreadcount);
        chkForUserAbort;
      end else begin
        {second phase, recover bad sectors}
        for sector:=1 to numsectors do begin
          {Did we read this sector successfully already?  Skip this sector if so}
          if readStatus[track,head,sector]=good
            then continue;
          DisplayDriveStatus(track,head,sector,false);
          {To pretty-up the display and make it easier to know if INT13 returned nothing}
          fillchar(sectorbuffer^,sectorsize,0);
          curstatus:=ReadSectors(CurDrive,head,track,sector,1,sectorbuffer);
          if ignoreController then curstatus:=$1c1c;
          {$IFDEF NODISKIO}
          curstatus:=fakestatus;
          curstatus:=fakestatus;
          move(ptr($40,$6c)^,sectorbuffer^,sectorsize);
          {$ENDIF}
          DisplayStatus('Last status code: '+DriveStatusString(hi(curstatus)));
          recoverCount:=0;
          if Reconstruct
            then fillchar(salvageData,sizeof(salvageData),0);
          if (Recover and (lo(curstatus)<>1)) or ignoreController then repeat
            inc(recoverCount);
            DisplayDriveStatus(track,head,sector,true);
            {read directly into track buffer on the heap}
            curstatus:=ReadSectors(CurDrive,head,track,sector,1,sectorbuffer);
            if ignoreController then curstatus:=$1c1c;
            {$IFDEF NODISKIO}
            curstatus:=fakestatus;
            move(ptr($40,$6c)^,sectorbuffer^,sectorsize);
            {$ENDIF}
            {was this read good?  If so, stop retrying the read}
            if (lo(curstatus)=1)
              then continue;
            if Reconstruct then begin
              if recoverCount=1 then DisplayStatus('Attempting reconstruction');
              readStatus[track,head,sector]:=unknown;
              {copy into reconstruction buffer}
              move(sectorbuffer^,salvagedSector,sectorsize);
              {update reconstruction iteration count}
              salvageData[sectorsize*8]:=recoverCount;
              {update reconstruction stats}
              updateSalvageCounts;
              DisplaySector(salvagedSectorp,0);
            end else begin
              DisplayStatus('(Recover) Drive status: '+DriveStatusString(hi(curstatus))
                           +'  Re-Reading, attempt #'+inttostr(recoverCount));
              DisplaySector(sectorbuffer,0);
            end;
            {Reconstruction is an expensive operation at 4.77 MHz,
            and we also want the heads off the media as soon as possible...
            but it is neat to watch it refine real-time, so we'll
            reconstruct the sector and user display every 4th read.}
            if (Reconstruct=true) and ((recoverCount AND 3)=0) then begin
              reconstructSector;
              DisplayStatus('Reconstructed sector from '+inttostr(recoverCount)+' samples...');
              DisplaySector(salvagedSectorp,salvageSecLine);
              tmps:='Reconstructed:';
              tfastwritexy(@tmps,vidp,0,salvageSecLine,normColor);
              {if the rebuilt sector stops changing, we can early abort}
              if BuffersEqual(salvagedSectorp,oldSectorp,sectorsize)<>0 then begin
                {copy current reconstructed buffer to storage}
                move(salvagedSector,oldSector,sizeof(oldSector));
                refineCount:=0;
              end else begin
                {they're equal -- make a note of it}
                inc(refineCount);
                {if they've been equal for a threshold, conclude reconstruction}
                if refineCount>=refineLimit
                  then break
                  else DisplayStatus('Reconstructed sector appears stable; concluding refinement in '
                       +inttostr(refineLimit-refineCount+1)+' iterations');
              end;
            end;
            {handle user input clumsily here}
            if (keypressed=true) then begin
              {empty the keyboard buffer}
              if keypressed then while keypressed do
                ch:=upcase(readkeychar);
              {if key was R then recalibrate the track,
              else assume user wants to abort this sector}
              if ch='R' then begin
                ch:=#0;
                status:=ResetDrive(CurDrive);
                DisplayStatus('Recalibrating drive and track...')
              end else begin
                if Reconstruct then begin
                  DisplayStatus('(Reconstruction) Finishing up...');
                  readStatus[track,head,sector]:=reconstructed;
                end else begin
                  DisplayStatus('(Recover) Discarding this sector...');
                  readStatus[track,head,sector]:=bad;
                end;
                break;
              end;
            end;
          until (lo(curStatus)=1) or (recoverCount>maxRecoverCount); {until 1 sector read or out of time}

          if (lo(curstatus)<>1) or ignoreController then begin
            {If we're here, we didn't read a sector successfully.}
            if (sector=9) then inc(badsec9s);
            if (head=(numheads-1)) then inc(badhead1s);
            {either invalidate the sector, or write the reconstruction}
            if Reconstruct then begin
              {perform final reconstruction}
              reconstructSector;
              move(salvagedSector,sectorbuffer^,sectorsize);
              CommitSector(sectorbuffer,track,head,sector);
              readStatus[track,head,sector]:=reconstructed;
              DisplayReadMatrix;
            end else begin
              DisplayStatus('Sector could not be recovered');
              if readStatus[track,head,sector]<>bad
                then fatalerror(90,'Bug: Bad sector isn''t actually bad');
            end;
          end else begin
            {We got a good read.  Update the status matrix and commit sector}
            DisplayStatus('Read successful');
            {$IFDEF NODISKIO}
            {$IFDEF SANITY}
            fillchar(sectorbuffer^,sectorsize,'.');
            {$ENDIF}
            {$ENDIF}
            l:=abspos(track,head,sector);
            CommitSector(sectorbuffer,track,head,sector);
            readStatus[track,head,sector]:=good;
            DisplayReadMatrix;
          end;
          {Display whatever we ended up with}
          imageTemp^.seek(abspos(track,head,sector));
          imageTemp^.read(tempSector^,sectorsize);
          DisplaySector(tempSector,0);
          inc(sectorcount); {used for estimated time display}
          chkForUserAbort;
        end;
      end;
    end;
  end;

  {$IFDEF SANITY}
  for track:=0 to numtracks-1 do
    for head:=0 to numheads-1 do
      for sector:=1 to numsectors do begin
        fillchar(sectorbuffer^,sectorsize,(head SHL 4) OR (sector AND $0F) );
        CommitSector(sectorbuffer,track,head,sector);
      end;
  {$ENDIF}
  {$IFDEF DEBUG}
  imageTemp^.seek(0);
  if recover then begin
    for track:=0 to numtracks-1 do
      for head:=0 to numheads-1 do
        for sector:=1 to numsectors do begin
          imageTemp^.seek(abspos(track,head,sector));
          imageTemp^.read(b,1);
          if b=ord('!')
            then writeln('WTF at t:h:s ',track,':',head,':',sector);
        end;
  end;
  {$ENDIF}
  DisplayReadMatrix;

  {Restore previous DBT LSOT val}
  DBT^.LastSectorOnTrack:=oldLSOT;
  ResetDrive(CurDrive);
end;

Procedure DumpDone;

  procedure ReportError(actualtype,newswitch:string);
  begin
    writeln('The disk you just dumped appears to ',actualtype,'.');
    writeln('You should re-dump it with the "-',newswitch,'" command-line switch.');
    writeln;
  end;

const
  reportErrors:boolean=false;

var
  l:longint;
  _t,_h,_s:byte;
  logfname,logtext:string;
  logfile:text;

begin
  FlushImage;

  freemem(tempsector,sectorsize);
  freemem(sectorbuffer,sectorsize);
  freemem(trackbuffer,tracksize);
  doneScreen;
  writeln('Done.');

  {if number of bad sector 9's == the number of tracks*heads, then
   every single sector 9 didn't read--which means it's probably an 8-sector
   disk.  Warn user appropriately.}
  if badsec9s=(numtracks*numheads)
    then reporterror('have 8 sectors per track','s8');

  {if number of bad sectors on the second side == everything then
   every single sector on the second side didn't read--which means it's probably a
   single-sided disk.  Warn user appropriately.}
  if badhead1s=(numtracks*numsectors)
    then reporterror('be a single-sided disk','h1');

  if numErrors<>0 then begin
    Chan2SquarewaveOn(200);
    writeln(numErrors,' errors dumping this disk; consider re-dumping/deleting the image.');
    l:=ticksSinceMidnight+18; repeat until ticksSinceMidnight > l;
    Chan2SquarewaveOff;
  end;
  writeln;

  {report failed sectors}
  for _t:=0 to numTracks-1 do
    for _h:=0 to numHeads-1 do
      for _s:=1 to numSectors do
        if readStatus[_t,_h,_s] <> good
          then reportErrors:=true;

  if reportErrors then begin
    logfname:=basename(imageName)+'.log';
    assign(logfile,logfname);
    rewrite(logfile);
    for _t:=0 to numTracks-1 do
      for _h:=0 to numHeads-1 do
        for _s:=1 to numSectors do
          if readStatus[_t,_h,_s] <> good then begin
            logtext:='Track: '+inttostr(_t)
                    +' Head: '+inttostr(_h)
                    +' Sector: '+inttostr(_s)
                    +' unreadable';
            if readStatus[_t,_h,_s] = reconstructed
              then logtext:=logtext+'; reconstruction was attempted'
              else logtext:=logtext+'; dumped as '+inttostr(sectorsize)+ ' "!" bytes instead';
            writeln(logtext);
            writeln(logfile,logtext);
          end;
    close(logfile);
    writeln('These results were written to ',logfname,' for your review.');
  end;
end;

begin
  DumpInit;
  if not IgnoreController then begin
    DumpDo(initial);
    if numerrors>0
      then DumpDo(recovery);
  end else begin
    DumpDo(recovery);
  end;
  DumpDone;
end.
